# -*- coding: utf-8 -*-

import os
import abc
import numpy as np      # >=1.10.4
from geco_stat._version import __version__, __release__
from geco_stat._constants import __default_bitrate__
from geco_stat.Abstract import Factory
from geco_stat.Abstract import AbstUnionable
from geco_stat.Abstract import AbstractPlottable
from geco_stat.Abstract import HDF5_IO
from geco_stat.Report import AbstReport
from geco_stat.Data import AbstData
from geco_stat.Data import Statistics
from geco_stat.Data import Histogram
from geco_stat.Time import TimeIntervalSet
from geco_stat.Timeseries import Timeseries

def run_unit_tests():
    print('Testing class initializations.')
    Timeseries((16384,))
    TimeIntervalSet()
    Histogram()
    Statistics()

    # TODO: make a timeseries and then from there unit test everything else

    print('Testing TimeIntervalSet arithmetic.')
    ti = TimeIntervalSet
    assert ti([66,69]) + ti([67,72]) == ti([66,72]), "Union failing"
    assert ti([66,69]) * ti([67,72]) == ti([67,69]), "Intersection failing"
    assert ti([66,69]) + ti([70,72]) == ti([66,69,70,72]), "Union failing"
    assert ti([66,73]) - ti([67,72]) == ti([66,67,72,73]), "Complement failing"
    assert ti([66,73]) - ti([66,73]) == ti(), "Complement failing"
    # TODO: Add some more arithmetic assertions.

    print('Testing TimeIntervalSet frame time rounding.')
    assert ti([65,124]).round_to_frame_times() == ti([64, 128]), \
        "Rounding to frame times is failing"
    assert ti([64,128]).round_to_frame_times() == ti([64, 128]), \
        "Rounding to frame times is failing"
    assert ti([63,65,120,133]).round_to_frame_times() == ti([0,192]), \
        "Rounding to frame times is failing"

    print('Testing TimeIntervalSet length calculation.')
    assert ti([6400,6464]).combined_length() == 64, \
        'Time interval total length calculations are off'
    assert ti([0,4,6400,6466]).combined_length() == 70, \
        'Time interval total length calculations are off'

    print('Testing TimeIntervalSet splitting into frame files.')
    try:
        # this should raise a value error since the frame times are not rounded
        ti([66,68]).split_into_frame_file_intervals()
        raise AssertionError('Should not be able to split a time interval not '
                             'having round endpoints')
    except ValueError:
        pass

    print('Testing HDF5 file saving capabilities.')
    ex = {
        'name': 'stefan',
        'age':  np.int64(24),
        'fav_numbers': np.array([2,4,4.3]),
        'fav_tensors': {
            'levi_civita3d': np.array([
                [[0,0,0],[0,0,1],[0,-1,0]],
                [[0,0,-1],[0,0,0],[1,0,0]],
                [[0,1,0],[-1,0,0],[0,0,0]]
            ]),
            'kronecker2d': np.identity(3)
        }
    }
    AbstReport.__save_dict_to_hdf5__(
        ex, 'geco_statistics_test_hdf5_dict_example.hdf5')
    loaded = AbstReport.__load_dict_from_hdf5__(
        'geco_statistics_test_hdf5_dict_example.hdf5')
    os.remove('geco_statistics_test_hdf5_dict_example.hdf5')
    np.testing.assert_equal(loaded, ex)

    # TODO: Add in tests for creating time intervals from strings
    # TODO: Add in HDF5 save/load tests for all classes

    clean_up()
    print('Unit tests passed!')


def clean_up():
    """
    Clean up side-effects after unit and integration tests have been run.
    """
    if os.path.exists('geco_statistics_test_hdf5_dict_example.hdf5'):
        os.remove('geco_statistics_test_hdf5_dict_example.hdf5')




# -*- coding: utf-8 -*-

import bisect
import subprocess
import numpy as np      # >=1.10.4
from geco_stat._version import __version__
from geco_stat.Abstract import Factory
from geco_stat.Abstract import AbstUnionable
from geco_stat.Abstract import AbstractPlottable
from geco_stat.Abstract import HDF5_IO
from geco_stat.Exceptions import VersionException

# Inherit from HDF5_IO first in order to get an implemented clone method
class TimeIntervalSet(HDF5_IO,
                      # AbstractPlottable, TODO Make AbstractPlottable
                      AbstUnionable):
    """
    TimeIntervalSet

    A class for storing sets of half open intervals of the form

        [t1,t2) U [t3,t4) U...

    along with methods for manipulating time interval sets, including

    - union(otherIntervalSet)
    - intersection(otherIntervalSet)
    - complement_with_respect_to(otherIntervalSet)
    - no_overlap_union(otherIntervalSet)

    These methods do not modify the TimeIntervalSet instance to which they are
    bound. This makes it easy to play around with them without annoying and
    potentially dangerous side-effects.
    """

    def __init__(self, intervalSet=None, start=None, end=None):
        """
        If no time interval is given at initialization, the TimeIntervalSet
        begins empty. There are two ways to initialize it with a nonempty set
        of time intervals:

        1)  Provide an ordered numeric list (float, int, or mixed type are
            all fine) with even length as the intervalSet argument. The
            lists elements

            [s1, e1, s2, e2, ... sN, eN]

            are interpreted as forming the interval union

            [s1, e1) U [s2, e2) U ... U [sN, eN]

        2)  Provide numerical start and end values, start=s, end=e,
            corresponding to the beginning and end of a single interval
            that will comprise the new interval set:

            [s, e)
        """
        if type(intervalSet) == list or type(intervalSet) == np.ndarray:
            if len(intervalSet) % 2 != 0:
                raise ValueError('intervalSet set must have even length (equal '
                                 'starts and ends)')
            elif not np.array_equal(sorted(intervalSet), intervalSet):
                raise ValueError('intervalSet must be sorted')
            else:
                self._data = np.array([float(x) for x in intervalSet])
                self.remove_empty_sets()
                self.assert_self_consistent()
        elif (intervalSet is None and start is None and end is None or
                start == end):
            self._data = np.array([])
        elif start < end:
            self._data = np.array([float(start), float(end)])
            self.remove_empty_sets()
            self.assert_self_consistent()
        else:
            raise ValueError('Invalid combination of arguments. '
                             'See documentation.')

    def __union__(self, other):
        """
        Return the union of the current set of intervals with some other set.

        Returns a new TimeIntervalSet instance without modifying the input
        arguments.
        """
        if len(other) == 0:
            return self.clone()
        elif len(self) == 0:
            return other.clone()
        # iteratively union every interval in the other set into this set
        result = self.clone()
        for i in range(0, len(other)//2):
            # this part is (mostly)  shared between set algebra methods
            start  = other.to_ndarray()[2*i]
            end    = other.to_ndarray()[2*i + 1]
            bnds   = result.__left_and_right_bounds__(start, end)
            left   = bnds[0] % 2
            right  = bnds[1] % 2
            npcat  = np.concatenate # give this a shorter name
            # the conditional responses are unique to each set algebra method
            if left == 0 and right == 1:
                result._data = npcat((result.to_ndarray()[0:bnds[0]],
                                      [start, end],
                                      result.to_ndarray()[bnds[1]+1:]))
            elif left == 0 and right == 0:
                result._data = npcat((result.to_ndarray()[0:bnds[0]],
                                      [start],
                                      result.to_ndarray()[bnds[1]+1:]))
            elif left == 1 and right == 1:
                result._data = npcat((result.to_ndarray()[0:bnds[0]],
                                      [end],
                                      result.to_ndarray()[bnds[1]+1:]))
            elif left == 1 and right == 0:
                result._data = npcat((result.to_ndarray()[0:bnds[0]],
                                      result.to_ndarray()[bnds[1]+1:]))
            result.remove_empty_sets()
        return result

    def to_ndarray(self):
        """
        Return a numpy.ndarray whose value can be used as an argument to this
        class's constructor to recreate an equivalent instance. For example,

        >>> TimeIntervalSet(numpy.array([0,1])).to_ndarray()
        array([ 0.,  1.])
        """
        return self._data

    @staticmethod
    def __find_frame_file_gps_start_time__(gps_time):
        """
        Get the GPS Time representing the START time of the frame file
        containing data for the time represented by the argument, which
        must also be in GPS Time format.

        Frame files always start at times that are integer multiples of
        64, so this is just a convenience function for rounding down to
        the nearest multiple of 64.
        """
        ans = np.floor(gps_time / 64.) * 64
        if ans % 1 != 0:
            raise ValueError("Out of precision in floats, answer should be "
                             "an integer")
        return ans

    @staticmethod
    def __find_frame_file_gps_end_time__(gps_time):
        """
        Get the GPS Time representing the END time of the frame file
        containing data for the time represented by the argument, which
        must also be in GPS Time format.

        Frame files always start at times that are integer multiples of
        64, so this is just a convenience function for rounding up to
        the nearest multiple of 64.
        """
        ans = np.ceil(gps_time / 64.) * 64
        if ans % 1 != 0:
            raise ValueError("Out of precision in floats, answer should "
                             "be an integer")
        return ans

    @staticmethod
    def tconvert(input_time):
        """
        Take either a numerical input (representing gps time) or a string input
        (representing UTC time) and return the opposite. For example:

            tconvert('Oct 30 00:00:00 GMT 2015')

        returns

            1130198417

        and

            tconvert(1130198417)

        returns

            'Oct 30 00:00:00 GMT 2015'

        """
        if type(input_time) is str:
            dump = subprocess.Popen(
                ["lalapps_tconvert",str(input_time)], stdout=subprocess.PIPE)
            converted_time = dump.communicate()[0]
            return int(converted_time) # should be an int; if not, that's bad
        else:
            dump = subprocess.Popen(["lalapps_tconvert",str(
                int(input_time))], stdout=subprocess.PIPE)
            converted_time = dump.communicate()[0]
            return converted_time

    def round_to_frame_times(self):
        """
        Return a TimeIntervalSet that is a superset of of this TimeIntervalSet
        and which perfectly overlaps with the data contained in a set of frame
        files. Since frame files start at times that are integer multiples of,
        this is tantamount to rounding the start times up and the end times
        down.
        """
        cls = type(self)
        rounded_times = cls()
        for i in range(0, len(self)//2):
            rounded_times += cls([
                cls.__find_frame_file_gps_start_time__(self.to_ndarray()[2*i]),
                cls.__find_frame_file_gps_end_time__(self.to_ndarray()[2*i+1])
            ])
        return rounded_times

    def split_into_frame_file_intervals(self):
        """
        Return a list of TimeIntervalSets corresponding to time intervals
        covered by the frame files covering this time range. For example,

        >>> TSet = geco_stat.TimeIntervalSet
        >>> TSet([64,192,256,320]).split_into_frame_file_intervals()
        [geco_stat.TimeIntervalSet([64.0, 128.0]), geco_stat.TimeIntervalSet([128.0, 192.0]), geco_stat.TimeIntervalSet([256.0, 320.0])]

        The input TimeIntervalSet instance must start and end on a valid
        frame file time (an integer multiple of 64) or else an error will
        be raised.
        """
        if self.round_to_frame_times() != self:
            raise ValueError("Can only split a rounded time interval")
        frame_intervals = []
        for i in range(0, len(self)//2):
            if int(self.to_ndarray()[2*i]) != self.to_ndarray()[2*i]:
                raise ValueError("Out of precision in floats, answer should "
                                 "be an integer")
            if int(self.to_ndarray()[2*i+1]) != self.to_ndarray()[2*i+1]:
                raise ValueError("Out of precision in floats, answer should "
                                 "be an integer")
            for start_time in range(int(self.to_ndarray()[2*i]),
                                    int(self.to_ndarray()[2*i+1]), 64):
                frame_intervals.append(type(self)([start_time, start_time+64]))
        return frame_intervals

    @classmethod
    def from_human_readable_strings(cls, readable_string_list):
        """
        Take an iterable consisting of pairs of strings and return a timeseries
        corresponding to those times. The input iterable should be flat
        (one-dimensional) and have an even number of entries, or else an error
        will result. For example, use as input:

            ['Oct 30 00:00:00 GMT 2015', 'Oct 30 00:02:00 GMT 2015']

        which will return a time interval corresponding to the given time
        strings. The conversion to GPS is carried out by lalapps_tconvert,
        so formatting must be comprehensible to that program.

        Note that no rounding occurs, so if you are going to use these times
        to find gravitational wave frame files, you should not use the resulting
        numerical values but find another way to generate a list of
        gravitational wave frame file start times (like using
        self.round_to_frame_times).
        """
        times = []
        for s in readable_string_list:
            if not isinstance(s, str):
                raise ValueError('from_human_readable_strings() must use '
                                 'strings as input')
            times.append(cls.tconvert(s))
        return cls(times)

    def intersection(self, other):
        """
        Return the intersection of the current set of intervals with some other
        set.

        Returns a new TimeIntervalSet instance without modifying the input
        arguments.
        """
        self.assert_self_consistent()
        other.assert_self_consistent()
        if len(other) == 0 or len(self) == 0:
            return TimeIntervalSet()
        result = TimeIntervalSet()
        for i in range(0, len(other)//2):
            # this part is (mostly)  shared between set algebra methods
            start  = other.to_ndarray()[2*i]
            end    = other.to_ndarray()[2*i + 1]
            bnds   = self.__left_and_right_bounds__(start, end) # this differs
            left   = bnds[0] % 2
            right  = bnds[1] % 2
            npcat  = np.concatenate # give this a shorter name
            # the conditional responses are unique to each set algebra method
            if left == 0 and right == 1:
                result._data = npcat((result.to_ndarray(),
                                      self.to_ndarray()[bnds[0]:bnds[1]+1]))
            elif left == 0 and right == 0:
                result._data = npcat((result.to_ndarray(),
                                      self.to_ndarray()[bnds[0]:bnds[1]+1],
                                      [end]))
            elif left == 1 and right == 1:
                result._data = npcat((result.to_ndarray(), [start],
                                      self.to_ndarray()[bnds[0]:bnds[1]+1]))
            elif left == 1 and right == 0:
                result._data = npcat((result.to_ndarray(), [start],
                                      self.to_ndarray()[bnds[0]:bnds[1]+1],
                                      [end]))
            result.remove_empty_sets()
        return result

    def complement_with_respect_to(self, other):
        """
        Return the complement of the current set of intervals with respect to
        another set. The other set must be a superset of the current set, or
        else an error will be raised.

        Returns a new TimeIntervalSet instance without modifying the input
        arguments.
        """
        self.assert_self_consistent()
        other.assert_self_consistent()
        if self.union(other) != other:
            raise ValueError('Can only take complement with respect to a '
                             'superset.')
        if len(self) == 0:
            return other
        result = TimeIntervalSet()
        for i in range(0, len(other)//2):
            # this part is (mostly)  shared between set algebra methods
            start  = other.to_ndarray()[2*i]
            end    = other.to_ndarray()[2*i + 1]
            bnds   = self.__left_and_right_bounds__(start, end) # this differs
            left   = bnds[0] % 2
            right  = bnds[1] % 2
            npcat  = np.concatenate # give this a shorter name
            # the conditional responses are unique to each set algebra method
            if left == 0 and right == 1:
                result._data = npcat((result.to_ndarray(), [start],
                                      self.to_ndarray()[bnds[0]:bnds[1]+1],
                                      [end]))
            elif left == 0 and right == 0:
                result._data = npcat((result.to_ndarray(), [start],
                                      self.to_ndarray()[bnds[0]:bnds[1]+1]))
            elif left == 1 and right == 1:
                result._data = npcat((result.to_ndarray(),
                                      self.to_ndarray()[bnds[0]:bnds[1]+1],
                                      [end]))
            elif left == 1 and right == 0:
                result._data = npcat((result.to_ndarray(),
                                      self.to_ndarray()[bnds[0]:bnds[1]+1]))
            result.remove_empty_sets()
        return result

    def __left_and_right_bounds__(self, a, b):
        """
        Used when merging an individual interval [a, b) into an existing
        TimeIntervalSet S. S must, implicitly, be sorted, of even length, and
        contain numerical data.

        This function returns a list L of length 2, such that:

            L[0] = index of leftmost value in S greater than or equal to a
            L[1] = index of rightmost value in S less than or equal to a

        L is then used to determine a merge strategy for the union,
        intersection, or complement of S with respect to [a, b).

        For the sake of simplicity of implementation, this function only accepts
        positive even length lists of start and end points. In other words, an
        error will be raised if the user tries passing an empty TimeIntervalSet.
        """
        if len(self.to_ndarray()) == 0:
            raise ValueError('Cannot use an empty TimeIntervalSet.')
        l = bisect.bisect_left(self.to_ndarray(), a)
        r = bisect.bisect_right(self.to_ndarray(), b) - 1
        return [l, r]

    def remove_empty_sets(self):
        """
        Find repeated endpoints and remove them pairwise. For example, the
        set

            [a, b) U [b, c)

        can be written

            [a, c)

        and the set

            [b, b)

        is empty, and can simply be removed.
        """
        self.assert_self_consistent() # check
        i = 0
        while i < len(self.to_ndarray()) - 1:
            if self.to_ndarray()[i] == self.to_ndarray()[i+1]:
                self._data = np.delete(self.to_ndarray(), np.s_[i:i+2])
            else:
                i += 1           # not a copy, move on to the next one

    def assert_unionable(self, other):
        if type(self) != type(other):
            raise ValueError('Type mismatch: cannot union ' +
                             str(type(self)) + ' with ' + str(type(other)))
        if self.__version__ != other.__version__:
            raise ValueError('TimeIntervalSets have different versions')
        return True

    def assert_self_consistent(self):
        'Check that this instance has form consistent with the class spec'
        if type(self.to_ndarray()) != np.ndarray:
            raise Exception('TimeIntervalSet corrupted: '
                            'data not a numpy.ndarray')
        elif not np.array_equal(sorted(self.to_ndarray()), self.to_ndarray()):
            raise Exception('TimeIntervalSet corrupted: data not sorted')
        elif len(self.to_ndarray()) % 2 != 0:
            raise Exception('TimeIntervalSet corrupted: odd number '
                            'of endpoints')
        return True

    def __clone__(self):
        return type(self)(self.to_ndarray())

    def combined_length(self):
        'Get the combined length of all time intervals in this TimeIntervalSet.'
        if len(self.to_ndarray()) == 0:
            return 0
        starts = self.to_ndarray()[0::2]
        ends   = self.to_ndarray()[1::2]
        length = 0
        for i in range(0, len(starts)):
            length += ends[i] - starts[i]
        return length

    def human_readable_dates(self):
        """
        Print the contained time intervals in an immediately human-readable
        form, assuming that the time endpoints that comprise this instance
        are GPS times. For example,

            [1135825217, 1135825219) U [1135825220, 1135825222)

        will be printed in the more comprehensible form:

            [Sun Jan 03 03:00:00 GMT 2016, Sun Jan 03 03:00:02 GMT 2016) U
            [Sun Jan 03 03:00:03 GMT 2016, Sun Jan 03 03:00:05 GMT 2016)

        """
        times = [str(int(x)) for x in self.to_ndarray()]
        self.assert_self_consistent()
        tstring = ""
        i = 0
        for time in times:
            dump = subprocess.Popen(
                ["lalapps_tconvert",time], stdout=subprocess.PIPE)
            if i == 0:
                tstring += '[' + dump.communicate()[0][:-1] + ', '
                i = 1
            else:
                tstring += dump.communicate()[0][:-1] + ') U\n'
                i = 0
        tstring = tstring[:-3] # shave off the last U character and newline
        return tstring

    @classmethod
    def __from_dict__(cls, d):
        return cls(d['data'])

    def __to_dict__(self):
        return {'data': np.array(self.to_ndarray())}

    def __eq__(self, other):
        return np.array_equal(self.to_ndarray(), other.to_ndarray())

    def __len__(self):
        return len(self.to_ndarray())

    def __mul__(self, other):
        'Multiplication can be used as a shorthand for intersection.'
        return type(self).intersection(self, other)

    def __sub__(self, other):
        """
        Subtraction can be used as a shorthand for complement. Specifically:

            a - b = b.complement_with_respect_to(a)

        Take careful note that the order of operations is switched in order
        for the notation to make sense. With this convention, it is easy to
        see that

            (a - b) + b = a - (b - b) = a - b + b = a

        so that it is manifestly associative, though it is non-commutative
        and doesn't obey proper group behavior since e.g. (-a) is not
        defined.
        """
        return type(self).complement_with_respect_to(other, self)

    def __str__(self):
        'Return a string expressing the object in set union notation'
        self.assert_self_consistent()
        if len(self) == 0:
            return '{}'
        starts = self.to_ndarray()[0::2]
        ends   = self.to_ndarray()[1::2]
        string = '[' + str(starts[0]) + ', ' + str(ends[0]) + ')'
        for i in range(1, len(starts)):
            string += ' U [' + str(starts[i]) + ', ' + str(ends[i]) + ')'
        return string

    def __repr__(self):
        self.assert_self_consistent()
        return __name__ + '.TimeIntervalSet(' + repr(list(self._data)) + ')'

    # TODO this shouldn't circularly Timeseries class... not elegant
    def from_timeseries(self, timeseries):
        """Just clone the TimeIntervalSet belonging to the Timeseries"""
        return timeseries.time_intervals.clone()

Factory.add_class(TimeIntervalSet)



# -*- coding: utf-8 -*-

import numpy as np      # >=1.10.4
from geco_stat._version import __version__, __release__
from geco_stat._constants import __default_bitrate__
from geco_stat.Exceptions import MissingChannelDataException
from geco_stat.Abstract import Factory
from geco_stat.Abstract import AbstUnionable
from geco_stat.Abstract import AbstractPlottable
from geco_stat.Abstract import HDF5_IO
from geco_stat.Time import TimeIntervalSet

# Inherit from HDF5_IO first in order to get an implemented clone method
class ReportSet(HDF5_IO,
                # AbstractPlottable, TODO Make AbstractPlottable
                AbstUnionable):
    """
    Class for collections of Reports, allowing for more advanced procedures
    that allow the user to distinguish between anomalous and typical time
    ranges in the input data.
    """

    # TODO Add notes, full intended time, current work block, and is_finished
    # method
    def __init__(self,
                 report_class_name,
                 bitrate                 = __default_bitrate__,
                 channel_name            = "blank_report",
                 time_intervals          = None,
                 report                  = None,
                 report_anomalies_only   = None,
                 report_sans_anomalies   = None,
                 missing_times           = None):

        if isinstance(report_class_name, str):
            self.report_class_name = report_class_name
            if not self.get_report_class() is type:
                raise ValueError('report_class must be equal to the name of '
                                 'a ReportData class')
        else:
            raise ValueError('report_class must be a string')

        if time_intervals is None:
            self.time_intervals         = TimeIntervalSet()
        else:
            self.time_intervals         = time_intervals.clone()

        if missing_times is None:
            self.missing_times          = TimeIntervalSet()
        else:
            self.missing_times          = missing_times.clone()

        # All or none of the three reports must be provided as arguments,
        # otherwise it would be possible to initialize an inconsistent
        # ReportSet.
        if (report is None and
                report_anomalies_only is None and
                report_sans_anomalies is None):
            self.report                 = \
                self.get_report_class()(bitrate=bitrate)
            self.report_anomalies_only  = \
                self.get_report_class()(bitrate=bitrate)
            self.report_sans_anomalies  = \
                self.get_report_class()(bitrate=bitrate)
        else:
            self.report                 = report.clone()
            self.report_anomalies_only  = report_anomalies_only.clone()
            self.report_sans_anomalies  = report_sans_anomalies.clone()

        assert np.int64(bitrate) == bitrate, 'bitrate must be an integer'
        self.bitrate                = np.int64(bitrate)
        self.channel_name           = channel_name

        self.assert_self_consistent()

    def get_report_class(self):
        """
        Get the specific class of report used in this report set.

        If the function is called from ReportSet, the user must provide the
        self argument, which is then either interpreted as an instance of
        ReportSet, or as the class name string itself. In the latter case,
        this function simply returns the Report class corresponding to the
        class name passed as an argument.
        """
        if not isinstance(self, str):
            self = self.report_class_name
        return globals()[self]

    @staticmethod
    def from_time_and_channel_name(report_class_name, channel_name,
                                   time_intervals, bitrate=__default_bitrate__):
        """
        Each subclass of ReportSet should have its own well-defined
        constructor that rejects initialization data that would lead to an
        instance un-unionable with a new blank instance of that subclass.

        This can be confirmed by checking at the end of initialization that the
        new instance is unionable with a new blank instance.

        The time_intervals argument must, at the moment, correspond to a single
        gravitational wave frame file. Future implementations might change this.
        """
        report_class = globals()[report_class_name]
        try:
            timeseries = Timeseries.from_time_and_channel_name(
                channel_name, time_intervals)
            missing_times = TimeIntervalSet()
            # FIXME: This line is broken. This is a staticmethod, no self.
            report = report_class.from_timeseries(self, timeseries)

            if report_class.is_anomalous(timeseries):
                report_anomalies_only = report
                report_sans_anomalies = report_class(
                    bitrate=bitrate, time_intervals=time_intervals)
            else:
                report_sans_anomalies = report
                report_anomalies_only = report_class(
                    bitrate=bitrate, time_intervals=time_intervals)
        except MissingChannelDataException:
            missing_times = time_intervals
            report = report_class(
                bitrate=bitrate,
                time_intervals=time_intervals)
            report_anomalies_only = report
            report_sans_anomalies = report

        # FIXME: This line is ALSO broken. This is a staticmethod, no cls.
        return cls(
            report_class_name       = report_class_name,
            bitrate                 = bitrate,
            channel_name            = channel_name,
            time_intervals          = time_intervals,
            report                  = report,
            report_sans_anomalies   = report_sans_anomalies,
            report_anomalies_only   = report_anomalies_only,
            missing_times           = missing_times
        )

    def assert_self_consistent(self):
        # TODO: make sure the r.__name__ business below works
        for r in (self.report, self.report_anomalies_only,
                  self.report_sans_anomalies):
            if not isinstance(r, self.get_report_class()):
                raise ValueError(
                    'key ' +
                    r.__name__ +
                    ' must be instance of ' +
                    self.report_class_name)
            r.assert_self_consistent()
            # r.assert_unionable(self.get_report_class()(self.bitrate))
            if self.bitrate != r.bitrate:
                raise ValueError(
                    'key ' +
                    r.__name__ +
                    ' has different bitrate than this ReportSet')
            if self.__version__ != r.__version__:
                raise ValueError(
                    'key ' +
                    r.__name__ +
                    ' has different version than this ReportSet')
        for t in self.time_intervals, self.missing_times:
            if not isinstance(t, TimeIntervalSet):
                raise ValueError(
                    'key ' +
                    t.__name__ +
                    ' must be instance of TimeIntervalSet')
            t.assert_self_consistent()
            if self.__version__ != t.__version__:
                raise ValueError(
                    'key ' +
                    t.__name__ +
                    ' has different version than this ReportSet')
        if (self.report_anomalies_only + self.report_sans_anomalies !=
                self.report):
            raise ValueError(
                'whole report should be union of anomalous and nominal parts')
        if self.missing_times + self.time_intervals != self.time_intervals:
            raise ValueError(
                'missing times should be subset of all times in ReportSet')
        if self.time_intervals != self.report.time_intervals:
            raise ValueError(
                'time intervals in full Report and ReportSet should match')
        assert np.int64(
            self.bitrate) == self.bitrate, 'bitrate must be an integer'

    def assert_unionable(self, other):
        if not isinstance(self, type(other)):
            raise ValueError('instances of ReportSet must be of same type')
        if self.get_report_class() != other.get_report_class():
            raise ValueError(
                'instances of ReportSet must have same Report class')
        if self.channel_name != other.channel_name:
            raise ValueError(
                'instances of ReportSet must have same channel_name')
        if self.bitrate != other.bitrate:
            raise ValueError('instances of ReportSet must have same bitrate')
        if self.__version__ != other.__version__:
            raise ValueError('instances of ReportSet must have same version')
        if self.time_intervals.intersection(
                other.time_intervals) == TimeIntervalSet([]):
            raise ValueError('instances of ReportSet cannot cover overlapping '
                             'time intervals')

    def __union__(self, other):
        ans = self.clone()
        ans.time_intervals          += other.time_intervals
        ans.missing_times           += other.missing_times
        ans.report                  += other.report
        ans.report_anomalies_only   += other.report_anomalies_only
        ans.report_sans_anomalies   += other.report_sans_anomalies

    def __clone__(self):
        return type(self)(
            report_class_name       = self.report_class_name,
            bitrate                 = self.bitrate,
            channel_name            = self.channel_name,
            time_intervals          = self.time_intervals,
            report                  = self.report,
            report_anomalies_only   = self.report_anomalies_only,
            report_sans_anomalies   = self.report_sans_anomalies,
            missing_times           = self.missing_times
        )

    @classmethod
    def __from_dict__(cls, d):
        return cls(
            report_class_name       = d['report_class_name'],
            bitrate                 = d['bitrate'],
            channel_name            = d['channel_name'],
            time_intervals          = TimeIntervalSet.from_dict(
                d['time_intervals']),
            report                  = cls.get_report_class(
                d['report_class_name']).from_dict(d['report']),
            report_anomalies_only   = cls.get_report_class(
                d['report_class_name']).from_dict(
                d['report_anomalies_only']),
            report_sans_anomalies   = cls.get_report_class(
                d['report_class_name']).from_dict(
                d['report_sans_anomalies']),
            missing_times           = TimeIntervalSet.from_dict(
                d['missing_times'])
        )

    def __to_dict__(self):
        return {
            'report_class_name':      self.report_class_name,
            'bitrate':                np.int64(self.bitrate),
            'channel_name':           self.channel_name,
            'time_intervals':         self.time_intervals.to_dict(),
            'report':                 self.report.to_dict(),
            'report_anomalies_only':  self.report_anomalies_only.to_dict(),
            'report_sans_anomalies':  self.report_sans_anomalies.to_dict(),
            'missing_times':          self.missing_times.to_dict()
        }

    def __eq__(self, other):
        try:
            self.assert_self_consistent()
            other.assert_self_consistent()
        except ValueError():
            return False
        if not isinstance(self, type(other)):
            return False
        if self.report_class_name != other.report_class_name:
            return False
        if self.bitrate != other.bitrate:
            return False
        if self.__version__ != other.__version__:
            return False
        if self.channel_name != other.channel_name:
            return False
        if self.time_intervals != other.time_intervals:
            return False
        if self.missing_times != other.missing_times:
            return False
        if self.report != other.report:
            return False
        if self.report_anomalies_only != other.report_anomalies_only:
            return False
        if self.report_sans_anomalies != other.report_sans_anomalies:
            return False
        return True




# -*- coding: utf-8 -*-

import abc
import numpy as np      # >=1.10.4
from geco_stat._version import __version__
from geco_stat._constants import __default_bitrate__
from geco_stat.Abstract import Factory
from geco_stat.Abstract import AbstUnionable
from geco_stat.Abstract import AbstractPlottable
from geco_stat.Abstract import HDF5_IO
from geco_stat.Time import TimeIntervalSet

# Inherit from HDF5_IO first in order to get an implemented clone method
class AbstData(HDF5_IO,
                         # AbstractPlottable, TODO Make AbstractPlottable
                         AbstUnionable):
    """
    Abstract class for data that can be aggregated. Subclasses should represent
    small, conceptually unified collections of information. For example,
    a subclass of AbstData might focus exclusively on gathering
    basic statistics, like the mean and median values of a periodic timeseries.
    
    As of this writing, AbstData DOES NOT necessarily store time
    information, since time information might not be necessary for every
    single type of ReportData. This is very much subject to change.
    """
    __metaclass__  = abc.ABCMeta
    __version__ = __version__

# TODO: Make AbstractPlottable
class Histogram(AbstData):
    """
    A class for storing a histogram of (quasi) periodic timeseries. In
    the current version, each period is assumed to last for one second.
    Bitrate is adjustable. The idea is to compare many periods worth
    of a signal in order to determine how much the signal varies over
    the entire timespan considered.

    This class DOES NOT contain information about the time ranges
    included in the data; that information should go into some containing
    class. This class is intended as a diagnostic report primitive.
    """

    def __init__(self,
                 hist            = None,
                 hist_range      = (-1e3, 1e3),
                 hist_num_bins   = 256,
                 bitrate         = __default_bitrate__):
        """
        Initialize an instance of the class. All properties have default
        values corresponding to an empty statistics set; they can be
        individually overridden.
        """
        # make sure hist_range is an ordered pair of numbers
        if not len(hist_range) == 2:
            raise ValueError('second argument (hist_range) must have length 2')
        elif hist_range[0] >= hist_range[1]:
            raise ValueError('min val of hist bin range must be less than max')

        # set values to "empty" histograms
        if hist is None:
            hist = np.zeros((hist_num_bins, bitrate), dtype=np.int64)

        assert np.int64(hist_num_bins) == hist_num_bins, \
            'hist_num_bins must be an integer'
        self.hist_num_bins  = np.int64(hist_num_bins)
        assert len(hist_range) == 2
        self.hist_range     = np.array(hist_range)
        # Make sure this is a copy of the data
        self.hist           = np.array(hist, copy=True)
        self.hist_bins      = np.linspace(hist_range[0], hist_range[1],
                                          hist_num_bins+1)
        self.t_ticks        = np.linspace(0,1,bitrate+1)
        assert np.int64(bitrate) == bitrate, 'bitrate must be an integer'
        self.bitrate        = np.int64(bitrate)

    def __union__(self, other):
        """
        Take the union of these two histograms, representing the histogram of
        the union of the two histograms' respective datasets.
        """
        ans         = self.clone()
        ans.hist    = self.hist + other.hist
        return ans

    def __clone__(self):
        return type(self)(
            hist            = self.hist,
            hist_range      = self.hist_range,
            hist_num_bins   = self.hist_num_bins,
            bitrate         = self.bitrate
        )

    def assert_unionable(self, other):
        if (self.hist_range != other.hist_range or
                self.hist_num_bins != other.hist_num_bins):
            raise ValueError('Histograms have different bin edges')
        if self.bitrate != other.bitrate:
            raise ValueError('Histograms have different bitrates')
        if self.__version__ != other.__version__:
            raise ValueError('Histograms have different versions')
        if not isinstance(self, type(other)):
            raise ValueError('Type mismatch: cannot union ' +
                             str(type(self)) + ' with ' + str(type(other)))
        return True

    def assert_self_consistent(self):
        if self.__version__ != __version__:
            raise ValueError(
                'Histogram version ' +
                self.__version__ +
                ' does not match lib version')
        assert self.hist_bins == np.linspace(
            self.hist_range[0], self.hist_range[1], self.hist_num_bins+1)
        assert self.t_ticks == np.linspace(0,1,self.bitrate+1)
        assert np.int64(
            self.bitrate) == self.bitrate, 'bitrate must be an integer'
        return True

    @classmethod
    def __from_dict__(cls, d):
        return cls(
            hist            = d['hist'],
            hist_range      = d['hist_range'],
            hist_num_bins   = d['hist_num_bins'],
            bitrate         = d['bitrate']
        )

    def __to_dict__(self):
        return {
            'hist': np.array(self.hist),
            'hist_range': np.array(self.hist_range),
            'hist_num_bins': self.hist_num_bins,
            'bitrate': self.bitrate,
            'version': self.__version__,
            'class': 'Histogram'
        }

    def from_timeseries(self, timeseries):
        num_sec = timeseries.get_num_seconds()
        hist, xedges, yedges = np.histogram2d(
            timeseries.flatten(),
            np.tile(self.t_ticks, num_sec),
            bins = [self.hist_bins, self.t_ticks])
        assert xedges == self.hist_bins, \
            'xedges should be the histogram bins'
        assert yedges == self.t_ticks, \
            'yedges should be the t_ticks'
        assert self.bitrate == timeseries.bitrate, \
            'timeseries and histogram must have same bitrate'
        self.assert_self_consistent()
        return type(self)(
            hist            = hist,
            hist_range      = self.hist_range,
            hist_num_bins   = self.hist_num_bins,
            bitrate         = self.bitrate)

    def __eq__(self, other):
        if (self.hist_range != other.hist_range or
                self.hist_num_bins != other.hist_num_bins):
            return False
        if self.bitrate != other.bitrate:
            return False
        if self.__version__ != other.__version__:
            return False
        if not isinstance(self, type(other)):
            return False
        return np.array_equal(self.hist, other.hist)

# TODO: Make AbstractPlottable
class Statistics(AbstData):
    """
    A class for storing diagnostic statistics for the aLIGO timing system.
    Includes methods for iteratively generating, amalgamating, and
    displaying statistics.

    This class DOES NOT contain information about the time ranges
    included in the data; that information should go into some containing
    class. This class is intended as a diagnostic report primitive.
    """

    def __init__(self,
                 sum             = None,
                 sum_sq          = None,
                 max             = None,
                 min             = None,
                 num             = 0,
                 bitrate         = __default_bitrate__):
        """
        All properties have default values corresponding to an empty statistics
        set; they can be individually overridden.
        """
        # set values of sum, sum_sq, and the histograms, since these depend on
        # bitrate and hist_num_bins and hence cannot be set above
        if sum is None:
            sum        = np.zeros(bitrate)
        if sum_sq is None:
            sum_sq     = np.zeros(bitrate)
        if max is None:
            # lowest possible max, cannot survive
            max        = np.ones(bitrate) * np.finfo(np.float64).min
        if min is None:
            min        = np.ones(bitrate) * \
                np.finfo(np.float64).max # same for min

        self.sum        = np.array(sum, copy=True).reshape((bitrate,))
        self.sum_sq     = np.array(sum_sq, copy=True).reshape((bitrate,))
        self.max        = np.array(max, copy=True).reshape((bitrate,))
        self.min        = np.array(min, copy=True).reshape((bitrate,))
        assert np.int64(
            num) == num, 'must provide an integer number of previous seconds'
        self.num        = np.int64(num)
        assert np.int64(bitrate) == bitrate
        self.bitrate    = np.int64(bitrate)

        self.assert_self_consistent()

    def __union__(self, other):
        """
        Take the union of these statistics, representing the same statistics
        taken on the union of the two statistics objects' respective datasets.
        """
        ans         = self.clone()
        ans.sum     = self.sum      + other.sum
        ans.sum_sq  = self.sum_sq   + other.sum_sq
        ans.max     = self.max      + other.max
        ans.min     = self.min      + other.min
        ans.num     = self.num      + other.num
        return ans

    def __clone__(self):
        self.assert_self_consistent()
        return type(self)(
            sum             = self.sum,
            sum_sq          = self.sum_sq,
            max             = self.max,
            min             = self.min,
            num             = self.num,
            bitrate         = self.bitrate
        )

    def assert_unionable(self, other):
        if self.bitrate != other.bitrate:
            raise ValueError('Statistics have different bitrates')
        if self.__version__ != other.__version__:
            raise ValueError('Statistics have different versions')
        if not isinstance(self, type(other)):
            raise ValueError('Type mismatch: cannot union ' +
                             str(type(self)) + ' with ' + str(type(other)))
        return True

    def assert_self_consistent(self):
        assert self.sum.shape == (self.bitrate,), \
            "sum should be vector with length equal to bitrate"
        assert self.sum_sq.shape == (self.bitrate,), \
            "sum_sq should be vector with length equal to bitrate"
        assert self.max.shape == (self.bitrate,), \
            "max should be vector with length equal to bitrate"
        assert self.min.shape == (self.bitrate,), \
            "min should be vector with length equal to bitrate"
        assert np.int64(self.num) == self.num
        assert np.int64(self.bitrate) == self.bitrate, \
            "bitrate must be an integer"
        if self.__version__ != __version__:
            raise ValueError(
                'Statistics version ' +
                self.__version__ +
                ' does not match lib version')
        if not ((self.bitrate,) == self.sum.shape ==
                self.sum_sq.shape == self.max.shape == self.min.shape):
            raise ValueError('Statistics fields must be 1-D with length equal '
                             'to bitrate')
        return True

    @classmethod
    def __from_dict__(cls, d):
        return cls(
            sum     = d['sum'],
            sum_sq  = d['sum_sq'],
            max     = d['max'],
            min     = d['min'],
            num     = d['num'],
            bitrate = d['bitrate']
        )

    def __to_dict__(self):
        assert self.num == np.int64(self.num)
        return {
            'sum':      np.array(self.sum).flatten(),
            'sum_sq':   np.array(self.sum_sq).flatten(),
            'max':      np.array(self.max).flatten(),
            'min':      np.array(self.min).flatten(),
            'num':      np.int64(self.num),
            'bitrate':  np.int64(self.bitrate),
            'version':  self.__version__,
            'class':    'Statistics'
        }

    def from_timeseries(self, timeseries):
        self_type = type(self)
        return self_type(
            sum     = timeseries.sum(0),
            sum_sq  = np.power(timeseries, 2).sum(0),
            max     = np.max(timeseries, 0),
            min     = np.min(timeseries, 0),
            num     = timeseries.shape[0],
            bitrate = timeseries.bitrate)

    def __eq__(self, other):
        if self.bitrate != other.bitrate:
            return False
        if self.__version__ != other.__version__:
            return False
        if not isinstance(self, type(other)):
            return False
        return (
            np.array_equal(self.sum,    other.sum)      and
            np.array_equal(self.sum_sq, other.sum_sq)   and
            np.array_equal(self.max,    other.max)      and
            np.array_equal(self.min,    other.min)      and
            np.array_equal(self.num,    other.num)
        )

# TODO: Add a ReportData wrapper for slow channel timeseries, complete with
# plotting

Factory.add_class(Histogram)
Factory.add_class(Statistics)



# -*- coding: utf-8 -*-


class VersionException(Exception):
    """
    For when the user attempts to instantiate an object using a
    datastructure from a different version number of the class.
    """


class MissingChannelDataException(Exception):
    """
    For when an attempt has been made to fetch a timeseries for
    a given time interval, but no data has been found, or the
    data request has timed out.
    """



# -*- coding: utf-8 -*-

import abc
import numpy as np      # >=1.10.4
from geco_stat._version import __version__
from geco_stat._constants import __default_bitrate__
from geco_stat.Abstract import Factory
from geco_stat.Data import AbstData
from geco_stat.Time import TimeIntervalSet

# TODO: Make AbstractPlottable
class AbstReport(AbstData):
    """
    A class for generating reports on data integrity. Should be extended to
    create reports specific to different types of data, e.g. IRIGBReport
    and DuoToneReport.

    The AbstReport class contains information on the time intervals included
    as well as basic statistics (mean, max, min, standard deviation)
    on the time intervals included, and finally, multiple histograms
    covering multiple "zoom" levels, for a tailored view of the data.

    Arguments:
    bitrate         the bitrate of the signal considered. defaults to 16384.

    time_intervals  the time intervals over which the data used to create
                    the report were recorded. defaults to an empty time
                    interval.

    data            a dictionary containing ReportData objects. defaults to
                    an empty histogram and an empty statistics instance with
                    the same bitrate.
    """
    __metaclass__  = abc.ABCMeta

    def __init__(self,
                 bitrate         = __default_bitrate__,
                 time_intervals  = None,
                 data            = None):

        assert np.int64(bitrate) == bitrate, 'bitrate must be an integer'
        self.bitrate = np.int64(bitrate)
        if time_intervals is None:
            self.time_intervals = TimeIntervalSet()
        else:
            self.time_intervals = time_intervals.clone()

        if data is None:
            data = self.__report_data_prototype__(bitrate)
        self._data = data           # data grouped here
        for key in data:            # instance attr pointers for convenience
            if hasattr(self, key):
                raise ValueError('AbstData dictionary should not '
                                 'have attributes conflicting with '
                                 'AbstReport attributes.')
            setattr(self, key, data[key])
        self.assert_self_consistent()

    @classmethod
    @abc.abstractmethod
    def __report_data_prototype__(cls, bitrate=__default_bitrate__):
        """
        MUST BE A CLASSMETHOD.

        Returns an empty ReportData._data dictionary representative of this
        ReportSet subclass. The only information necessary should be the
        bitrate; all other configuration information for the ReportData
        instances should be specified as part of this function, and should be
        considered to be characteristic of this subclass definition.

        A good operating definition of this prototypical empty class is that
        it defines unionability.

        EXAMPLE:
        --------

        @classmethod
        def __report_data_prototype__(cls, bitrate=__default_bitrate__)
            data = {
                'histogram': Histogram(bitrate=bitrate),
                'statistics': Statistics(bitrate=bitrate)
            }
        """

    @staticmethod
    @abc.abstractmethod
    def is_anomalous(timeseries):
        """
        MUST BE A STATICMETHOD.

        Define a method for testing whether a timeseries is anomalous. If so,
        the report generated from this timeseries will be unioned into
        report_anomalies_only. If not, the report generated from this
        timeseries will be unioned into report_sans_anomalies. In any case,
        the report will be unioned into report, which contains report data on
        the entire timeseries contained in the ReportSet.
        """

    def fold_in_timeseries(self, timeseries, time_intervals,
                           bitrate=__default_bitrate__):
        """
        Return a new report containing the current report's data along with
        data gleaned from the timeseries provided as an argument folded in.
        """
        return self.union(type(self).from_timeseries(
            timeseries, time_intervals, bitrate))

    def __union__(self, other):
        ans = self.clone()
        ans.time_intervals  += other.time_intervals
        for key in ans._data:
            ans._data[key] += other._data[key]
        return ans

    def __clone__(self):
        cloned_data = self._data
        for key in cloned_data:
            cloned_data[key] = cloned_data[key].clone()
        return type(self)(
            bitrate         = self.bitrate,
            time_intervals  = self.time_intervals.clone(),
            data            = cloned_data
        )

    def assert_unionable(self, other):
        if self.bitrate != other.bitrate:
            raise ValueError('Reports have different bitrates')
        if self.__version__ != other.__version__:
            raise ValueError('Reports have different versions')
        if not isinstance(self, type(other)):
            raise ValueError('Type mismatch: cannot union ' +
                             str(type(self)) + ' with ' + str(type(other)))
        if set(self._data) != set(other._data):
            raise ValueError(
                'AbstData sets do not have matching key sets.')
        if self.time_intervals.intersection(
                other.time_intervals) != TimeIntervalSet():
            raise ValueError('Reports have overlapping time intervals.')
        return True

    # TODO confirm self data unionability with new class instance
    def assert_self_consistent(self):
        """
        Confirm that this Report is self-consistent. It should not generally
        be necessary to modify this, except perhaps to extend it in subclasses.
        """
        for key in self._data:
            if not isinstance(self._data[key], AbstData):
                raise ValueError(
                    'key ' +
                    str(key) +
                    ' must be instance of AbstData')
            self._data[key].assert_self_consistent()
            if self.bitrate != self._data[key].bitrate:
                raise ValueError('Report constituents have different bitrates')
            if self.__version__ != self._data[key].__version__:
                raise ValueError('Report constituents have different versions')
        if not isinstance(self.time_intervals, TimeIntervalSet):
            raise ValueError(
                'self.time_intervals must be an instance of TimeIntervalSet.')
        if self.__version__ != self.time_intervals.__version__:
            raise ValueError(
                'time_intervals has different version than the Report itself')
        assert np.int64(
            self.bitrate) == self.bitrate, 'bitrate must be an integer'

    # FIXME this needs to be completely revamped to reflect a simpler structure
    # for Reports. It seems that subclassing AbstReport should mainly just
    # involve specifying the classes of ReportData that it uses.
    @classmethod
    def __from_dict__(cls, d):
        data = dict()
        for key, value in d['data'].items():
            # FIXME implement this using the Factory pattern used elsewhere.
            report_data_class = globals()[ value['class'] ]
            if not issubclass(report_data_class, AbstData):
                raise ValueError('Cannot reconstruct Report data; class '
                                 'property not a valid AbstData '
                                 'subclass')
            data[key] = report_data_class.from_dict(value)
        return cls(
            bitrate         = d['bitrate'],
            time_intervals  = TimeIntervalSet.from_dict(
                d['time_intervals']),
            data            = data
        )

    def __to_dict__(self):
        data = dict()
        for key in self._data:
            data[key] = self._data[key].to_dict()
        return {
            'bitrate':          self.bitrate,
            'time_intervals':   self.time_intervals.to_dict(),
            'data':             data
        }

    def __eq__(self, other):
        if not isinstance(self, type(other)) or set(
                self._data) != set(other._data):
            return False
        if (self.bitrate != other.bitrate or
                self.__version__ != other.__version__):
            return False
        if self.time_intervals != other.time_intervals:
            return False
        for key in self._data:
            if self._data[key] != other._data[key]:
                return False
        return True

class IRIGBReport(AbstReport):
    def __init__(self):
        # TODO: Implement
        raise NotImplementedError()

    def is_anomalous(timeseries):
        # For now, always assume non-anomalous
        return False


class DuoToneReport(AbstReport):
    def __init__(self):
        # TODO: Implement
        raise NotImplementedError()

    def is_anomalous(timeseries):
        # For now, always assume non-anomalous
        return False

Factory.add_class(IRIGBReport)
Factory.add_class(DuoToneReport)



# -*- coding: utf-8 -*-

import os
import h5py             # >=2.5.0
import abc
import numpy as np      # >=1.10.4
from geco_stat._version import __version__
from geco_stat._constants import __default_bitrate__
from geco_stat.Exceptions import VersionException

# You can store classes in the factory for later recovery, but you have to
# manually add classes to the factory after defining them.
class Factory(object):
    """
    A factory class for mapping the names of AbstractDictRepresentable subclass
    names to their corresponding subclasses. Classes can be added to the
    factory using the ``add_class`` method and recovered using the
    ``get_class`` method, which is useful for, e.g., reconstructing an object
    from a dictionary when the class of that object is specified as a string.

    For example, if ``d`` is such a dictionary, then

        cls = geco_stat.Factory.get_class(d['class'])

    can be used to get (what is presumably) the original class associated with
    the instance represented by that dictionary. The rest of the data in that
    dictionary should then be sufficient (along with the class itself) to
    reconstruct the original object. At the time of writing, this is the
    approach favored by the ``__from_dict__`` private methods used by
    ``from_dict``.

    You can store classes in the factory for later recovery, but you have to
    manually add classes to the factory after defining them. This approach has
    the advantage of facilitating runtime modifications for prototyping or
    advanced munging/scripting purposes; simply defining a new class and
    adding it to the factory should be enough to give it all of the features of
    built-in geco_stat classes.
    """

    _factory = dict()

    @staticmethod
    def add_class(newcls, name=None):
        """
        Store a class in the factory for later retrieval.

        >>> geco_stat.Factory.add_class(object, 'object')
        """
        if name is None:
            name = newcls.__name__
        Factory._factory[name] = newcls 

    @staticmethod
    def get_class(clsname):
        """
        Retrieve a class by name:

        >>> geco_stat.Factory.get_class('object')
        object
        """
        return Factory._factory[clsname]

    @staticmethod
    def classes():
        """
        List all classes currently in the factory.

        >>> geco_stat.Factory.classes()
        [object]
        """
        return Factory._factory.values()

    @staticmethod
    def classnames():
        """
        List the names of all classes currently in the factory.

        >>> geco_stat.Factory.classnames()
        ['object']
        """
        return Factory._factory.keys()

    @staticmethod
    def items():
        """
        Return a list of key-value pair tuples containing the names
        and classes for each class stored in the factory.

        >>> geco_stat.Factory.items()
        [('object', object)]
        """
        return Factory._factory.items()

    @staticmethod
    def del_class(clsname):
        """
        Remove a class from the factory.

        >>> geco_stat.Factory.del_class('object')
        """
        del Factory._factory[clsname]

class AbstTimeIntervals(object):
    """
    The AbstTimeIntervals defines what objects can specify time intervals
    over which data might be collected. In practice, it should only have one
    implementation; the abstract class serves mainly to keep the type graph
    acyclic and restrictive.

    This abstract class is meant for organizational/conceptual purposes and
    does not specify an interface or implement any methods.
    """
    __metaclass__  = abc.ABCMeta
    __version__ = __version__

class AbstractTimeSeries(object):
    """
    The AbstractTimeSeries provides a way of defining the category of objects
    which can serve as valid sources of information for geco_stat. In
    practice, it should only have one implementation; the abstract class serves
    mainly to keep the type graph acyclic.

    This abstract class is meant for organizational/conceptual purposes and
    does not specify an interface or implement any methods.
    """
    __metaclass__  = abc.ABCMeta
    __version__ = __version__

class AbstTimeSeriesDerivable(object):
    """
    The purpose of ``geco_stat`` is to look at data and do things with it,
    so it is fitting that the primordial interface provides methods for
    creating a new instance using a Timeseries.
    """
    __metaclass__  = abc.ABCMeta
    __version__ = __version__

    @abc.abstractmethod
    def clone(self):
        """
        Create and return a new instance of this class which shares all
        values (and which would return true in an equality test) with this
        instance, but which shares no object pointers (even in sub-properties)
        and can hence be modified with impunity without fear of side-effects.
        """

    def get_times(self):
        """
        Get an AbstTimeIntervals instance representing the time intervals
        covered by this instance.
        """
        # If storing times differently than suggested here, implement a
        # different method. But using ``_times`` seems reasonable.
        if isinstance(self, AbstTimeIntervals):
            return self.clone()
        elif isinstance(self._times, AbstTimeIntervals):
            return self._times.clone()

    @classmethod
    @abc.abstractmethod
    def from_timeseries(cls, timeseries):
        """
        Create a compatible object using an AbstractTimeSeries as input.
        """

class AbstUnionable(AbstTimeSeriesDerivable):
    """
    An interface describing classes whose members contain information that is
    associated with specific time intervals. These members can be unioned or
    "added" together provided that they don't contain redundant time
    information, i.e., provided their time interval sets are non-intersecting.
    This makes it possible to iteratively build instances of AbstUnionable that
    cover large, possibly disjoint spans of time from smaller "atomic"
    instances simply by unioning.
    """
    __metaclass__  = abc.ABCMeta
    __version__ = __version__

    # Use a template method approach to always make sure that these items
    # are unionable before proceeding.
    def union(self, other):
        """
        Combine these two instances to cover a larger span of time. Like the
        ``clone`` method, this should return an instance sharing no object
        pointers with the original instances.
        """
        self.assert_self_consistent()
        other.assert_self_consistent()
        self.assert_unionable(other)
        return self.__union__(other)

    # There is no reason to check for consistency every time; too much
    # abstraction with no clarifying purpose.
    # FIXME deprecated
    @abc.abstractmethod
    def __union__(self, other):
        """
        Aggregate these two instances without first checking that the instances
        are compatible or self-consistent. This is part of the implementation
        of the union method.
        """


    @abc.abstractmethod
    def __eq__(self, other):
        """
        Instances must have a way of determining equality that is based on
        values rather than ids, i.e. if the contents of two instances contain
        the same information, then they are equal, regardless of whether
        they are the same instance in memory.
        """

    def __ne__(self, other):
        return not self == other


    def __add__(self, other):
        """
        Addition is a shorthand for union.
        """
        return self.union(other)

    # TODO The private method name is undesirable and should be
    # replaced with something better. Not deprecated -- for now.
    @abc.abstractmethod
    def assert_self_consistent(self):
        """
        Make sure this instance is self-consistent.
        """

    @abc.abstractmethod
    def assert_unionable(self, other):
        """
        Make sure these two instances can be unioned.
        """

class AbstractDictRepresentable(object):
    """
    Subclasses of AbstractDictRepresentable contain information that can be
    represented as a **PARTICULAR** type of dictionary. To be specific, they
    can be represented as dictionaries whose elements are one of:

    1. A string.
    2. A numpy.int64.
    3. A numpy.float64.
    4. A numpy.ndarray.
    5. A AbstractDictRepresentable instance.

    This allows for unambiguous file saving and loading through the uniform
    intermediate step of a dictionary representation.
    """
    __metaclass__  = abc.ABCMeta
    __version__ = __version__

    def to_dict(self):
        """
        Return a dictionary whose elements consist of strings, ints, lists, or
        numpy.ndarray objects, or of other dicts whose contents follow this
        pattern recursively. This dictionary must wholly represent the data in
        this object, so that this object may be totally reconstructed using
        the dictionary's contents. This is an implementation method used to
        store data in HDF5.
        """
        # Unless overridden, this method will always check version and class
        # information, with class-specific treatment of the dictionary being
        # handled by a template method, __to_dict__.
        d = self.__to_dict__()
        d['version'] = self.__version__
        d['class'] = type(self).__name__
        return d

    @staticmethod
    def from_dict(d):
        """
        Construct an instance of this class using a dictionary of the form
        output by self.to_dict. Should generally be a class method.
        """
        # Unless overridden, this method will always check version and class
        # information, with class-specific treatment of the dictionary being
        # handled by a template method, __from_dict__.
        cls = Factory.get_class(d['class'])
        if d['version'] != cls.__version__:
            raise VersionException("Tried constructing an instance of %s "
                                   "using old version %s"
                                   % (cls.__name__, d['version']))
        return cls.__from_dict__(d)

    def clone(self):
        """
        Create and return a new instance of this class which shares all
        values (and which would return true in an equality test) with this
        instance, but which shares no object pointers (even in sub-properties)
        and can hence be modified with impunity without fear of side-effects.
        """
        # If this implementation makes you nervous (for anything besides
        # performance reasons, then you are not properly implementing the
        # to_dict and from_dict methods.
        return self.from_dict(self.to_dict())

    @abc.abstractmethod
    def __to_dict__(self):
        """
        Return a dictionary whose elements consist of strings, ints, lists, or
        numpy.ndarray objects, or of other dicts whose contents follow this
        pattern recursively. This dictionary must wholly represent the data in
        this object, so that this object may be totally reconstructed using
        the dictionary's contents. This is an implementation method used to
        store data in HDF5.
        """

    # MUST BE A CLASS METHOD.
    @abc.abstractmethod
    def __from_dict__(cls, d):
        """
        Construct an instance of this class using a dictionary of the form
        output by self.to_dict. Should generally be a class method.
        """

class HDF5_IO(AbstractDictRepresentable):
    """
    Uses the ``to_dict`` methods of AbstractDictRepresentable to save and load
    HDF5 files of arbitrary complexity.
    """
    __metaclass__  = abc.ABCMeta
    __version__ = __version__

    def save_hdf5(self, filename):
        """Save this instance to an hdf5 file."""
        self.__save_dict_to_hdf5__(self.to_dict(), filename)

    @classmethod
    def load_hdf5(cls, filename):
        """Load an instance saved in an hdf5 file."""
        # TODO: Make this a AbstReport-defined staticmethod that only
        # needs class information that is already provided in the saved
        # dictionary.
        return cls.from_dict(cls.__load_dict_from_hdf5__(filename))

    @classmethod
    def __save_dict_to_hdf5__(cls, dic, filename):
        """
        Save a dictionary whose contents are only strings, np.float64,
        np.int64, np.ndarray, and other dictionaries following this structure
        to an HDF5 file. These are the sorts of dictionaries that are meant
        to be produced by the AbstractDictRepresentable.to_dict() method. The saved
        dictionary can then be loaded using __load_dict_to_hdf5__(), and the
        contents of the loaded dictionary will be the same as those of the
        original.
        """
        if os.path.exists(filename):
            raise ValueError('File %s exists, will not overwrite.' % filename)
        with h5py.File(filename, 'w') as h5file:
            cls.__recursively_save_dict_contents_to_group__(h5file, '/', dic)

    @classmethod
    def __recursively_save_dict_contents_to_group__(cls, h5file, path, dic):
        """
        Take an already open HDF5 file and insert the contents of a dictionary
        at the current path location. Can call itself recursively to fill
        out HDF5 files with the contents of a dictionary.
        """
        # argument type checking
        if not isinstance(dic, dict):
            raise ValueError("must provide a dictionary")
        if not isinstance(path, str):
            raise ValueError("path must be a string")
        if not isinstance(h5file, h5py._hl.files.File):
            raise ValueError("must be an open h5py file")
        # save items to the hdf5 file
        for key, item in dic.items():
            if not isinstance(key, str):
                raise ValueError("dict keys must be strings to save to hdf5")
            # save strings, numpy.int64, and numpy.float64 types
            if isinstance(item, (np.int64, np.float64, str)):
                h5file[path + key] = item
                if not h5file[path + key].value == item:
                    raise ValueError('The data representation in the HDF5 file '
                                     'does not match the original dict.')
            # save numpy arrays
            elif isinstance(item, np.ndarray):
                h5file[path + key] = item
                if not np.array_equal(h5file[path + key].value, item):
                    raise ValueError('The data representation in the HDF5 file '
                                     'does not match the original dict.')
            # save dictionaries
            elif isinstance(item, dict):
                cls.__recursively_save_dict_contents_to_group__(
                    h5file, path + key + '/', item)
            # other types cannot be saved and will result in an error
            else:
                raise ValueError('Cannot save %s type.' % type(item))

    @classmethod
    def __load_dict_from_hdf5__(cls, filename):
        """
        Load a dictionary whose contents are only strings, floats, ints,
        numpy arrays, and other dictionaries following this structure
        from an HDF5 file. These dictionaries can then be used to reconstruct
        AbstractDictRepresentable subclass instances using the
        AbstractDictRepresentable.from_dict() method.
        """
        with h5py.File(filename, 'r') as h5file:
            return cls.__recursively_load_dict_contents_from_group__(h5file,
                                                                     '/')

    @classmethod
    def __recursively_load_dict_contents_from_group__(cls, h5file, path):
        """
        Load contents of an HDF5 group. If further groups are encountered,
        treat them like dicts and continue to load them recursively.
        """
        ans = {}
        for key, item in h5file[path].items():
            if isinstance(item, h5py._hl.dataset.Dataset):
                ans[key] = item.value
            elif isinstance(item, h5py._hl.group.Group):
                ans[key] = cls.__recursively_load_dict_contents_from_group__(
                    h5file, path + key + '/')
        return ans

class AbstractPlottable(object):
    """
    An interface for generating matplotlib figures that can be used in
    visualizing data.
    """
    __metaclass__  = abc.ABCMeta

    @abc.abstractmethod
    def plot(self):
        """
        Create some sort of visualization for the information content of this
        object. Container objects should use this to
        recursively call plotting functions in their constituents and generate
        summary plots representing all their information.
        """

    @abc.abstractmethod
    def summary(self):
        """
        Create some sort of verbose, human-readable text summary for the
        information content of this object. Should return a string.
        """




# -*- coding: utf-8 -*-

import os
import subprocess
import numpy as np      # >=1.10.4
from geco_stat._constants import __default_bitrate__
from geco_stat.Exceptions import MissingChannelDataException
from geco_stat.Time import TimeIntervalSet


class Timeseries(np.ndarray):
    """
    A thin wrapper for ndarrays, adding in a couple of convenience methods
    used to create new instances from gravitational wave frame files.
    """

    @classmethod
    def from_time_and_channel_name(
        cls,
        channel_name,
        time_interval,
        bitrate=__default_bitrate__
    ):
        """
        Load a timeseries using this channel_name and time_interval.

        The time_interval argument must, at the moment, correspond to a single
        gravitational wave frame file. Future implementations might change this.
        """
        frame_path = cls.locate_frame_file(channel_name, time_interval)
        return cls.from_frame_file(
            channel_name,
            frame_path,
            time_interval,
            bitrate)

    @classmethod
    def from_frame_file(
        cls,
        channel_name,
        path,
        time_intervals,
        bitrate=__default_bitrate__
    ):
        """
        Load channel from file path to array.

        If a channel doesn't exist within the
        frame file located at the specified path, this function will return
        None. Otherwise, it returns an ndarray with one row for each second of
        data in the frame file.
        """

        # make sure path exists
        if not os.path.exists(path):
            raise ValueError('Path does not exist: ' + path)

        # set up the processes for acquiring and processing the data
        dump = subprocess.Popen(
            ["framecpp_dump_channel","--channel",channel_name,path],
            stdout=subprocess.PIPE)
        data_string = dump.communicate()[0]
        # print(now() + ' Timeseries retrieved, beginning processing.')

        # remove headers from the data
        formatted_data_string = cls.__remove_header_and_text__(data_string)

        # if the string is empty, the channel didn't exist. return None.
        if formatted_data_string == '':
            raise MissingChannelDataException()

        # instantiate numpy array and return it; will have number of rows equal
        # to the number of seconds in a frame file and number of columns equal
        # to the bitrate of the channel.
        ans = np.fromstring(formatted_data_string, sep=',').reshape(
            (64, bitrate)).view(cls)
        ans.time_intervals = time_intervals
        ans.bitrate = bitrate
        return ans

    def get_num_seconds(self):
        """
        Get the number of seconds worth of data in this frame file.
        """
        if self.shape[0] != self.time_intervals.combined_length():
            raise Exception('data corrupted; mismatch between time_interval '
                            'length and number of seconds of data')
        return self.shape[0]

    @staticmethod
    def locate_frame_file(channel_name, time_interval):
        """
        Find the gravitational wave frame file corresponding to a particular
        time interval for a particular channel name.
        """
        if time_interval.round_to_frame_times() != time_interval:
            raise ValueError('time interval must be rounded to frame times')
        if time_interval.combined_length() != 64:
            raise ValueError('time interval must fit a single frame file')
        if len(time_interval) != 2:
            raise ValueError('time interval set must be 1 continuous interval')
        if not isinstance(channel_name, str):
            raise ValueError('channel_name must be a string')
        if not isinstance(time_interval, TimeIntervalSet):
            raise ValueError('time_interval must be a TimeIntervalSet')

        detector_prefix = channel_name[0]
        dump = subprocess.Popen([
            'gw_data_find',
            '-o', detector_prefix,
            '-t', detector_prefix + '1_R',
            '-s', time_interval.to_ndarray()[0],
            '-e', time_interval.to_ndarray()[1],
            '-u', 'file'], stdout=subprocess.PIPE)
        frame_path = dump.communicate()[0]
        if frame_path[0:15] != 'file://localhost':
            raise Exception('expected file://localhost prefix output from ' +
                            'gw_data_find, got %s' % frame_path[0:15])
        frame_path = frame_path[16:]
        if not os.path.exists(frame_path):
            raise Exception('gw_data_find returned faulty ' +
                            'path:\n\t %s' % frame_path)
        return frame_path

    @staticmethod
    def __remove_lines__(string, num_lines):
        """
        remove first n lines of a string
        """
        i = 0
        n = 0
        l = len(string)
        while n < num_lines:
            i = string.find('\n', i+1)
            # if a newline isn't found, this means there are no new lines left.
            if i == -1:
                return ""
            # if the string ends on a newline, return empty string.
            elif i+1 == l:
                return ""
            n += 1
        return string[i+1:]

    @classmethod
    def __remove_header_and_text__(cls, string):
        """
        Delete first 6 lines; all spaces; and the word 'Data:' which precedes
        the numerical data.

        (This replaces sed and tr in the original implementation with native
        python.)
        """
        return cls.__remove_lines__(string, 6).translate(None, 'Dat: ')
